import mne

def mark_bad_chs(data, process_params):
    data_bads = data.copy()
    # clear bads of any existing channels to avoid duplicates
    data_bads.info["bads"] = []
    for ch in process_params["bads"]:
        data_bads.info["bads"].append(ch)

    return data_bads

def interp_bad_chs(data):
    print("channels to interpolate: ", data.info["bads"])
    data_bads_interp = data.copy().interpolate_bads(reset_bads=False, verbose=True)
    return data_bads_interp

# ---- test -----------------------------------------------------

# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# epochs = mne.Epochs(raw, events, verbose=False)
# evoked = epochs.average()
#
# process_params = {"bads": ['MEG 2443', 'EEG 053'], "interpolate": True}
#
# def test(data, process_params):
#     data_bad_chs = mark_bad_chs(data, process_params)
#     if process_params["interpolate"]:
#         data_bads_interp = interp_bad_chs(data_bad_chs)
#         print("bad channels interpolation finished")
#         return(data_bads_interp)
#
#     else:
#         print("bad channels finished")
#         return data_bad_chs
#
#
# output = test(raw, process_params)
# print("bads: ", output.info["bads"])
# print(output.info["ch_names"])
# print(type(output))

# to test interpolation for raw data:
# raw.plot(butterfly=True, color='#00000022', bad_color='r')
# output.plot(butterfly=True, color='#00000022', bad_color='r', block=True)




import sys

from PyQt5.QtWidgets import QApplication

from main_window import MainWindow

if __name__ == '__main__':
    app = QApplication(sys.argv)
    main_window = MainWindow(app)
    main_window.show()
    sys.exit(app.exec_())



def baseline_func(data, process_params):
    epochs_baselined = data.copy().apply_baseline(process_params["interval"])
    return epochs_baselined


# ---- test -----------------------------------------------------
#
# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# epochs = mne.Epochs(raw, events, verbose=False)
# # evoked = epochs.average() # even though a baseline was not applied (and should be None by default), there was no difference between the epochs.average() with and without the baseline
#
# # read in evoked data that has not been baselined
# sample_data_evk_file = os.path.join(sample_data_folder, 'MEG', 'sample',
#                                     'sample_audvis-ave.fif')
# right_vis = mne.read_evokeds(sample_data_evk_file, condition='Right visual')
#
#
# process_params = {"interval": (-0.1, 0.0)}
#
# try:
#     output = baseline_func(epochs, process_params)
#     print(output.info)
#     print(type(output))
# except ValueError as VE:
#     print("Error: ", str(VE))
# except AttributeError as AE:
#     print("Error: Baseline can only be applied to epoched or evoked data")
#
# print("baselining finished")
#
# right_vis.plot()
# output.average().plot()
#






import mne

def filter_func(data, process_params):
    data_filtered = data.copy().filter(l_freq=process_params["l_freq"],
                                       h_freq=process_params["h_freq"],
                                       method=process_params["method"])
    return data_filtered


# ---- test -----------------------------------------------------

# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# epochs = mne.Epochs(raw, events, verbose=False)
# evoked = epochs.average()
#
# process_params = {"l_freq": 10, "h_freq":70, "method": "fir"}
#
# output = filter_func(raw, process_params)
# print(output.info)
# print(type(output))



import mne


def get_preset_pipeline(pipeline_name, data):

    # Need to get bads before running the pipeline
    example_pipeline = [    ["filter", {"l_freq": 1, "h_freq": 40, "method": "fir"}],
                            ["resample", {"sfreq": 500}],
                            ["line_noise", {"freqs": (50, 100), "method": "iir"}],
                            ["channels", {"bads": [], "interpolate": True}],
                            ["ICA", {"n_components": 20,
                                     "ica_method": "fastica",
                                     "extended": False,
                                     "autoreject": True,
                                     "rejection_method": "manual"}],
                            ["epoch", {"tmin": -0.2, "tmax": 0.5, "autoreject": True}],
                            ["baseline", {"interval": (-0.2, 0)}],
                            ["", {}],
                            ["", {}],
                            ["", {}]
                        ]

    # Need to get bads before running the pipeline
    gramfort_pipeline = [   ["", {}], # Spare one at the start for optional step eg resampling to be applied before pipeline is run
                            ["filter", {"l_freq": 1, "h_freq": 40, "method": "fir"}],
                            ["channels", {"bads": [], "interpolate": True}],
                            ["ICA", {"n_components": 0.99,
                                      "ica_method": "fastica",
                                      "extended": False,
                                      "autoreject": True,
                                      "rejection_method": "manual"}],
                            ["epoch", {"tmin": -0.2, "tmax": 0.5, "tstep": None, "autoreject": True}],
                            ["baseline", {"interval": (-0.2, 0)}],
                            ["", {}],
                            ["", {}],
                            ["", {}],
                            ["", {}]
                        ]

    if pipeline_name == "example":
        return example_pipeline
    elif pipeline_name == "gramfort":
        return gramfort_pipeline

    # following pipelines require rank to be calculated
    #rank_dict = mne.compute_rank(data, info=data.info, verbose=False)
    #rank = sum(rank_dict.values())

    # alternatively, just use number of channels - quicker to calculate, and chances are by the time some processing has been done
    # the rank will have reduced from what it was for the original raw data so it will have to be redone anyway
    picks = mne.pick_types(data.info, meg=True, eeg=True, eog=False, ecg=False)
    # print(len(picks))
    num_chs = len(picks)

    # Need to get bads and rank/num of chs before running the pipeline
    makoto_pipeline = [["filter", {"l_freq": 1, "h_freq": None, "method": "fir"}],
                       ["resample", {"sfreq": 250}],
                       ["channels", {"bads": [], "interpolate": True}],
                       ["set_reference", {"ref_channels": "average"}],
                       ["line_noise", {"freqs": (50), "method": "fir"}],
                       ["epoch", {"tmin": -0.2, "tmax": 0.5, "tstep": None, "autoreject": True}],
                       ["ICA", {"n_components": num_chs,
                                "ica_method": "infomax",
                                "extended": False,
                                "autoreject": False,
                                "rejection_method": "manual"}],
                       ["", {}],
                       ["", {}],
                       ["", {}]
                       ]

    # Need to get bads and rank/num of chs before running the pipeline
    luck_pipeline = [["", {}],  # blank for optional resampling
                     ["filter", {"l_freq": 0.1, "h_freq": None, "method": "fir"}],
                     ["line_noise", {"freqs": (50), "method": "fir"}],
                     ["channels", {"bads": [], "interpolate": True}],
                     ["ICA", {"n_components": num_chs,
                              "ica_method": "fastica",
                              "extended": False,
                              "autoreject": False,
                              "rejection_method": "manual"}],
                     ["set_reference", {"ref_channels": "average"}],
                     ["epoch", {"tmin": -0.2, "tmax": 0.5, "tstep": None, "autoreject": True}],
                     ["baseline", {"interval": (-0.2, 0)}],
                     ["", {}],
                     ["", {}]
                     ]

    # Need to get bads and rank/num of chs before running the pipeline
    resting_state_pipeline = [["filter", {"l_freq": 1, "h_freq": 100, "method": "iir"}],
                              ["line_noise", {"freqs": (50, 100), "method": "iir"}],  # check method
                              ["ICA", {"n_components": num_chs,
                                       "ica_method": "fastica",
                                       "extended": False,
                                       "autoreject": False,
                                       "rejection_method": "manual"}],
                              ["resample", {"sfreq": 256}],
                              ["epoch", {"tmin": None, "tmax": None, "tstep": 30, "autoreject": True}],
                              # ["baseline", {"interval": (-0.2, 0)}],
                              ["", {}],
                              ["", {}],
                              ["", {}],
                              ["", {}],
                              ["", {}]
                              ]

    if pipeline_name == "makoto":
        return makoto_pipeline
    elif pipeline_name == "luck":
        return luck_pipeline
    elif pipeline_name == "resting":
        return resting_state_pipeline


import mne
from autoreject import get_rejection_threshold


def epoch_func(data, process_params):
    if len(data.events) != 0: # events in data
        epochs = mne.Epochs(data, events=data.events, tmin=process_params["tmin"], tmax=process_params["tmax"])
    else: # no events in data
        events = mne.make_fixed_length_events(data, duration=process_params["tstep"])
        epochs = mne.Epochs(data, events=events, tmin=0.0, tmax=process_params["tstep"], baseline=(0, 0))
    return epochs


def drop_bad_epochs_func(data, process_params):

    include = []
    picks = mne.pick_types(data.info, meg=True, eeg=True, stim=False,
                           eog=True, include=include, exclude='bads')

    if len(data.events) != 0: # events in data

        events = data.events
        # Create epochs as advised in the autoreject docs to get thresholds.
        # (Note that detrending is applied and bads are excluded)
        threshold_epochs = mne.Epochs(data, events=events, tmin=process_params["tmin"],
                                      tmax=process_params["tmax"],
                                      picks=picks, baseline=(None, 0), preload=True,
                                      reject=None, verbose=True, detrend=1)
        # create actual epochs
        epochs = mne.Epochs(data, events=data.events, tmin=process_params["tmin"], tmax=process_params["tmax"])


    else: # no events in data

        # create events
        events = mne.make_fixed_length_events(data, duration=process_params["tstep"])
        # Create epochs as advised in the autoreject docs to get thresholds.
        threshold_epochs = mne.Epochs(data, events=events, tmin=0.0,
                                      tmax=process_params["tstep"],
                                      picks=picks, baseline=(0, 0), preload=True,
                                      reject=None, verbose=True, detrend=1)
        # create actual epochs
        epochs = mne.Epochs(data, events=events, tmin=0.0, tmax=process_params["tstep"], baseline=(0, 0))

    # get rejection thresholds based on recommended epochs (bad channels excluded)
    reject = get_rejection_threshold(threshold_epochs)  # removed decim=2 as RuntimeWarning: could cause aliasing artifacts

    # apply rejection thresholds to data with bad channels included
    epochs_drop_bad = epochs.copy().drop_bad(reject=reject)

    return epochs_drop_bad


# ---- test -----------------------------------------------------

# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# raw.events = []
# #raw.events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# #epochs = mne.Epochs(raw, raw.events, verbose=False)
# #evoked = epochs.average()
#
# def test(data):
#
#     process_params = {"tmin":None,
#                       "tmax": None,
#                       "tstep": 5.0,
#                       "autoreject": True}
#     data.stim_channel = None
#
#     if isinstance(data, mne.io.BaseRaw):
#         try:
#             if process_params["autoreject"]:
#                 epochs_drop_bad = drop_bad_epochs_func(data, process_params)
#                 epochs_drop_bad.plot()
#                 print("epochs_drop_bad")
#                 print(type(epochs_drop_bad))
#                 print(epochs_drop_bad.info["ch_names"])
#             else:
#                 epochs = epoch_func(data, process_params)
#                 epochs.plot()
#                 print("epochs")
#                 print(type(epochs))
#                 print(epochs.info["ch_names"])
#         except AttributeError as AE:
#             print("Error", "No stim channel selected")
#             return
#         except ValueError as VE:
#             print("Error: ", str(VE))
#             return
#     else:
#        print("Error: ", "Epoching can only be applied to raw data")
#
# test(raw)


import sys
from PyQt5.QtWidgets import QMessageBox, QApplication, QWidget


def alert_msg(window, title, msg):
    alert = QMessageBox(window)
    alert.setWindowTitle(title)
    alert.setText(msg)
    alert.exec_()

#test
#
# app = QApplication(sys.argv)
# main_window = QWidget()
# main_window.show()
# QMessageBox(main_window, "Alert", "message")
# sys.exit(app.exec_())



import mne

def resample_func(data, process_params):
    data_resampled = data.copy().resample(sfreq=process_params["sfreq"])
    #print(data_resampled.info)
    return data_resampled

def resample_data_and_events_func(data, process_params):
    data_resampled, events_resampled = data.copy().resample(sfreq=process_params["sfreq"], events=data.events)
    #print(data_resampled.info)
    return data_resampled, events_resampled



# ---- test -----------------------------------------------------
#
# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# epochs = mne.Epochs(raw, events, verbose=False)
# evoked = epochs.average()
#
# process_params = {"sfreq": 250}
#
# output = resample_func(raw, process_params)
# print(raw.info)
# print(output.info)
# print(type(output))



import mne
from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont

from mne.io.pick import channel_type, get_channel_type_constants

from prep_func_dialog_classes import options_dialog_cancel

from matplotlib import pyplot as plt

def check_for_picks_data(plot_window): # returns True if at least one set of picks_data present, else returns False
    if len(plot_window.picks_data_list[0]) == 0: # if picks_data_list[0] is an empty tuple, ie no channels were selected
        return False
    else:
        return True

class PickChannelsDialog(QWidget):
    def __init__(self, plot_window):
        super(PickChannelsDialog, self).__init__()

        self.plot_window = plot_window

        self.setWindowTitle("Pick Channels to Plot")

        info = self.plot_window.data_list[0][0].info

        channel_types = [k.upper() for k in get_channel_type_constants().keys()]

        self.channels_view = QListWidget()
        self.channels_view.setSpacing(7)
        self.channels_view.setSelectionMode(QListWidget.ExtendedSelection)

        for index, ch in enumerate(info["chs"]):
            kind = channel_type(info, index).upper()
            name = ch["ch_name"]
            item = QListWidgetItem(name + "    " + kind)
            self.channels_view.addItem(item)

            # if channel have already been picked, highlight them
            # Like 'saving' previously picked channels
            if hasattr(plot_window, "picks"):
                if name in plot_window.picks:
                    item.setSelected(True)

        self.update_btn = QPushButton('Update', self)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.plot_window))

        layout = QVBoxLayout()
        layout.addWidget(self.channels_view)
        layout.addWidget(self.update_btn)
        layout.addWidget(self.cancel_btn)
        self.setLayout(layout)

    def update(self):

        picks = []
        for item in self.channels_view.selectedItems():
            ch_name = item.text().split("   ")[0]
            picks.append(ch_name)

        if len(picks) > 0:
            self.plot_window.picks = picks
            self.plot_window.update_channels_for_picks()
            self.close()
        else:
            QMessageBox.warning(self, "No Channels Selected", "Please select at least one channel")


class TimeSeriesTab(QWidget):
    def __init__(self, plot_window):
        super(TimeSeriesTab, self).__init__()

        self.plot_window = plot_window

        self.plot_btn = QPushButton("Plot time series")
        self.plot_btn.clicked.connect(self.plot)

        layout = QVBoxLayout()
        layout.addWidget(self.plot_btn)
        self.setLayout(layout)

    def plot(self):

        if not check_for_picks_data(self.plot_window): # if picks data not present
            QMessageBox.warning(self, "Channels Not Selected", "Please select channels to plot")
            return

        else:
            for i in range(len(self.plot_window.picks_data_list)):

                if len(self.plot_window.picks_data_list[i]) != 0: # if data is present
                    data = self.plot_window.picks_data_list[i]
                    data_name = self.plot_window.data_list[i][1]
                    fig = data.plot(show=False)
                    fig.canvas.set_window_title(data_name)

            plt.show()

            #fig1 = self.plot_window.pick_data1.plot(show=False)
            #fig2 = self.plot_window.pick_data2.plot(show=False)
            #fig1.canvas.set_window_title(self.plot_window.data1_name)
            #fig2.canvas.set_window_title(self.plot_window.data2_name)

class PSDTab(QWidget):
    def __init__(self, plot_window):
        super(PSDTab, self).__init__()

        self.plot_window = plot_window

        self.average_check = QCheckBox("Plot average of channels")

        self.plot_btn = QPushButton("Plot pattern spectral density")
        self.plot_btn.clicked.connect(self.plot)

        layout = QVBoxLayout()
        layout.addWidget(self.average_check)
        layout.addWidget(self.plot_btn)
        self.setLayout(layout)

    def plot(self):

        if not check_for_picks_data(self.plot_window):  # if picks data not present
            QMessageBox.warning(self, "No Channels Selected", "Please select channels to plot")
            return

        else:
            average = self.average_check.isChecked()

            for i in range(len(self.plot_window.picks_data_list)):

                if len(self.plot_window.picks_data_list[i]) != 0:  # if data is present
                    data = self.plot_window.picks_data_list[i]
                    data_name = self.plot_window.data_list[i][1]
                    fig = data.plot_psd(show=False, average=average)
                    fig.canvas.set_window_title(data_name)

            plt.show()

            #fig1 = self.plot_window.pick_data1.plot_psd(show=False, average=average)
            #fig2 = self.plot_window.pick_data2.plot_psd(show=False, average=average)

            #fig1.canvas.set_window_title(self.plot_window.data1_name)
            #fig2.canvas.set_window_title(self.plot_window.data2_name)

class TopoEpochsEvokedTab(QWidget):
    def __init__(self, plot_window):
        super(TopoEpochsEvokedTab, self).__init__()

        self.plot_window = plot_window

        self.event_select = QComboBox()
        self.event_select_lbl1 = QLabel("Select event ID")
        self.event_select_lbl2 = QLabel("Only epochs/evoked responses containing this event \nwill be plotted")

        # bold font to use for titles:
        bold_font = QFont()
        bold_font.setBold(True)

        self.plot_epochs_title_lbl = QLabel("Epochs Image Map")
        self.plot_epochs_title_lbl.setFont(bold_font)
        self.plot_epochs_info_lbl = QLabel("A topographical map showing one image for each sensor.\n"
                                           "Each image shows all the epochs for that sensor, with each\n"
                                           "row of pixels representing a single epoch and colour scale\n"
                                           "respresenting the signal value at each time point.")
        self.plot_epochs_btn = QPushButton("Plot epoch images")
        self.plot_epochs_btn.clicked.connect(self.plot_epochs)

        self.plot_evoked_title_lbl = QLabel("Evoked Responses Map")
        self.plot_evoked_title_lbl.setFont(bold_font)
        self.plot_evoked_info_lbl = QLabel("A topographical map showing one image for each sensor.\n"
                                           "Each image shows the evoked response at that sensor.\n"
                                           "If two data files have been loaded,their equivalent\n"
                                           "evoked responses are compared on each plot.\n"
                                           "Note: plotting with 'All channels' or 'MEG only' selected\n "
                                           "will plot MEG channels. Select 'EEG only' to plot EEG channels")
        self.plot_evoked_btn = QPushButton("Plot evoked responses")
        self.plot_evoked_btn.clicked.connect(self.plot_evoked)

        layout = QGridLayout()
        layout.addWidget(self.event_select_lbl1, 0, 0)
        layout.addWidget(self.event_select, 0, 1, Qt.AlignLeft)
        layout.addWidget(self.event_select_lbl2, 1, 0, 1, 2)
        layout.addWidget(self.plot_epochs_title_lbl, 2, 0, 1, 2)
        layout.addWidget(self.plot_epochs_info_lbl, 3, 0, 1, 2)
        layout.addWidget(self.plot_epochs_btn, 4, 0, 1, 2)
        layout.addWidget(self.plot_evoked_title_lbl, 5, 0, 1, 2)
        layout.addWidget(self.plot_evoked_info_lbl, 6, 0, 1, 2)
        layout.addWidget(self.plot_evoked_btn, 7, 0, 1, 2)
        self.setLayout(layout)

    def populate_cbox(self, id_lst):
        self.event_select.clear()
        self.event_select.addItems(id_lst)

    def plot_epochs(self):

        if not check_for_picks_data(self.plot_window):  # if picks data not present
            QMessageBox.warning(self, "No Channels Selected", "Please select channels to plot")
            return

        else:

            try:
                event_id = self.event_select.currentText()

                for i in range(len(self.plot_window.picks_data_list)):

                    if len(self.plot_window.picks_data_list[i]) != 0:  # if data is present
                        data = self.plot_window.picks_data_list[i]
                        data_name = self.plot_window.data_list[i][1]


                        if "mag" in data:
                            layout = mne.channels.find_layout(data.info, ch_type='mag')
                            title = data_name + ": Magnetometers, Event ID: " + event_id
                            fig = data[event_id].plot_topo_image(layout=layout, fig_facecolor='w',
                                                                   font_color='k', sigma=1, title=title, show=False)
                            fig.canvas.set_window_title(data_name)

                        if "grad" in data:
                            layout = mne.channels.find_layout(data.info, ch_type='grad')
                            title = data_name + ": Gradiometers, Event ID: " + event_id
                            fig = data[event_id].plot_topo_image(layout=layout, fig_facecolor='w',
                                                                  font_color='k', sigma=1, title=title, show=False)
                            fig.canvas.set_window_title(data_name)
                        if "eeg" in data:
                            layout = mne.channels.find_layout(data.info, ch_type='eeg')
                            title = data_name + ": Electrodes, Event ID: " + event_id
                            fig = data[event_id].plot_topo_image(layout=layout, fig_facecolor='w',
                                                                  font_color='k', sigma=1, title=title, show=False)
                            fig.canvas.set_window_title(data_name)

                plt.show()

            except KeyError: # if event_is not present in data2
                QMessageBox.warning(self, "Event Error", "Selected event is not present in both data files")
                return

    def plot_evoked(self):

        if not check_for_picks_data(self.plot_window):  # if picks data not present
            QMessageBox.warning(self, "No Channels Selected", "Please select channels to plot")
            return

        else:
            event_id = self.event_select.currentText()

            evokeds = [] # empty list - populate with evoked data to plot

            for i in range(len(self.plot_window.picks_data_list)):

                if len(self.plot_window.picks_data_list[i]) != 0:  # if data is present

                    data = self.plot_window.picks_data_list[i]
                    data_name = self.plot_window.data_list[i][1]

                    event_dict = data.event_id
                    data_ids = [i for i in event_dict.keys()]

                    if event_id in data_ids: # check that the event ID is present in each data file

                        evoked = data.average()
                        evoked.comment = data_name + " - Event ID " + event_id

                        evokeds.append(evoked)
                    else:
                        QMessageBox.warning(self, "Event Error", "Selected event is not present in both data files")
                        return

            mne.viz.plot_evoked_topo(evokeds)
            #mne.viz.plot_compare_evokeds(evokeds, axes="topo")






import mne
import re
import sys

from PyQt5.QtWidgets import *



def get_valid_save_name(main_window):
    # get filename to save file under
    options = QFileDialog.Options()
    options |= QFileDialog.DontUseNativeDialog
    fpath, _ = QFileDialog.getSaveFileName(main_window, "Save As...", "", "", options=options)
    # validate fname
    if fpath:
        fname = fpath.split("/")[-1]

        # check for invalid characters (alphanumeric and underscores only)
        invalid_char = re.findall(r"[\W\s]", fname)
        if len(invalid_char) > 0:
            print(invalid_char)
            QMessageBox.warning(main_window, "Invalid File Name", "Files names should only include letters, digits, and underscores")
            return -1

        else:
            return fpath

def write_fif(data, fpath):

    if isinstance(data, mne.io.BaseRaw):
        save_fname = fpath + "-raw.fif"
        data.save(save_fname)
        return True
    elif isinstance(data, mne.BaseEpochs):
        save_fname = fpath + "-epo.fif"
        data.save(save_fname)
        return True
    elif type(data) == mne.Evoked or type(data) == mne.EvokedArray:
        save_fname = fpath + "-epo.fif"
        data.save(save_fname)
        return True
    else:
        return False


# --- test -------------------------------------------------------------------

if __name__ == '__main__':
    app = QApplication(sys.argv)
    main_window = QWidget()
    main_window.show()
    #data = 1
    fname = get_valid_save_name(main_window)
    if fname:
        print(fname)
        #write_fif(data, fname, main_window)
    sys.exit(app.exec_())




import mne
from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt

from mne.io.pick import channel_type, get_channel_type_constants

from msgbxs import alert_msg


def options_dialog_cancel(dialog_window, main_window):
    reply = QMessageBox.question(dialog_window, 'Confirm',
                                       "Are you sure you want to close this window? Any unsaved changes will be lost",
                                       QMessageBox.Ok | QMessageBox.Cancel, QMessageBox.Cancel)
    if reply == QMessageBox.Ok:
        dialog_window.close()
    else: # reply == QMessageBox.Cancel
        return

class FilterOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(FilterOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        # Then set up the layout
        self.setWindowTitle("Filter")

        self.method_lbl = QLabel("Filter type/method:")
        self.fir_rbtn = QRadioButton("FIR")
        self.iir_rbtn = QRadioButton("IIR")

        self.l_input_lbl = QLabel("Low cutoff frequency(Hz):")
        self.l_input = QLineEdit(self)
        self.l_input.setFixedWidth(60)
        self.h_input_lbl = QLabel("High cutoff frequency(Hz):")
        self.h_input = QLineEdit(self)
        self.h_input.setFixedWidth(60)

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        # if filter options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "filter":
            if process_params["method"] == "fir":
                self.fir_rbtn.setChecked(True)
                self.iir_rbtn.setChecked(False)
            elif process_params["method"] == "iir":
                self.fir_rbtn.setChecked(False)
                self.iir_rbtn.setChecked(True)
            if process_params["l_freq"] == None:
                self.l_input.setText("")
            else:
                self.l_input.setText(str(process_params["l_freq"]))
            if process_params["h_freq"] == None:
                self.h_input.setText("")
            else:
                self.h_input.setText(str(process_params["h_freq"]))
        else: # if setting filter options for the first time
            self.fir_rbtn.setChecked(True) # default value


        layout = QGridLayout()
        layout.addWidget(self.method_lbl, 0, 0, 1, 2)
        layout.addWidget(self.fir_rbtn, 1, 0)
        layout.addWidget(self.iir_rbtn, 2, 0)
        layout.addWidget(self.l_input_lbl, 3, 0)
        layout.addWidget(self.l_input, 3, 1)
        layout.addWidget(self.h_input_lbl, 4, 0)
        layout.addWidget(self.h_input, 4, 1)
        layout.addWidget(self.update_btn, 5, 0, 1, 2, Qt.AlignHCenter)
        layout.addWidget(self.cancel_btn, 6, 0, 1, 2, Qt.AlignHCenter)
        self.setLayout(layout)

    def update(self):

        for r_btn in [self.fir_rbtn, self.iir_rbtn]:
            if r_btn.isChecked():
                method = r_btn.text().lower()

        if self.l_input.text() == "":
            l_cutoff = None
        else:
            try:
                l_cutoff = int(self.l_input.text())
            except ValueError:
                QMessageBox.warning(self, "Invalid Input", "Low cutoff frequency must be an integer")
                self.l_input.clear()
                return

        if self.h_input.text() == "":
            h_cutoff = None
        else:
            try:
                h_cutoff = int(self.h_input.text())
            except ValueError:
                QMessageBox.warning(self, "Invalid Input", "High cutoff frequency must be an integer")
                self.h_input.clear()
                return


        process_name = "filter"
        process_params = {"l_freq": l_cutoff, "h_freq": h_cutoff, "method": method}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful","Filter settings updated")
        self.close()

class ResampleOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(ResampleOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        self.setWindowTitle("Resample")

        self.sfreq_spinbox_lbl = QLabel("New sampling frequency (Hz):")
        self.sfreq_spinbox = QSpinBox(self)
        self.sfreq_spinbox.setMaximum(9999)
        self.sfreq_spinbox.setFixedWidth(60)

        # if resample options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "resample":
            self.sfreq_spinbox.setValue(process_params["sfreq"])
        else: # if setting filter options for the first time
            self.sfreq_spinbox.setValue(250)  # default value

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.clicked.connect(self.update)
        self.update_btn.setFixedWidth(150)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        layout = QGridLayout()
        layout.addWidget(self.sfreq_spinbox_lbl, 0, 0)
        layout.addWidget(self.sfreq_spinbox, 0, 1)
        layout.addWidget(self.update_btn, 1, 0, 1, 2, Qt.AlignHCenter)
        layout.addWidget(self.cancel_btn, 2, 0, 1, 2, Qt.AlignHCenter)
        self.setLayout(layout)

    def update(self):

        process_name = "resample"
        process_params = {"sfreq": int(self.sfreq_spinbox.value())}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "Resampling settings updated")

        self.close()

class LineNoiseOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(LineNoiseOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        # Then set up the layout
        self.setWindowTitle("Remove Line Noise")

        self.method_lbl = QLabel("Notch filter type/method:")
        self.fir_rbtn = QRadioButton("FIR")
        self.iir_rbtn = QRadioButton("IIR")

        self.l_noise_lbl1 = QLabel("Line noise frequency(Hz):")
        self.l_noise_lbl2 = QLabel("(Harmonics can be included\neg 50, 100, 150)")
        self.l_noise_input = QLineEdit(self)
        self.l_noise_input.setFixedWidth(100)

        # if notch filter options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "line_noise":
            if process_params["method"] == "fir":
                self.fir_rbtn.setChecked(True)
                self.iir_rbtn.setChecked(False)
            elif process_params["method"] == "iir":
                self.fir_rbtn.setChecked(False)
                self.iir_rbtn.setChecked(True)
            if process_params["freqs"] == None:
                self.l_input.setText("")
            else:
                if type(process_params["freqs"]) == int: # just one value eg 50
                    freqs_str = str(process_params["freqs"])
                else: # tuple eg (50, 100)
                    freqs_str = str(process_params["freqs"])[1:-1] # [1:-1] to get rid of the tuple brackets
                self.l_noise_input.setText(freqs_str)
        else:  # if setting options for the first time (default values)
            self.fir_rbtn.setChecked(True)
            self.iir_rbtn.setChecked(False)
            self.l_noise_input.setText("50")

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        layout = QGridLayout()
        layout.addWidget(self.method_lbl, 0, 0, 1, 2)
        layout.addWidget(self.fir_rbtn, 1, 0)
        layout.addWidget(self.iir_rbtn, 2, 0)
        layout.addWidget(self.l_noise_lbl1, 3, 0)
        layout.addWidget(self.l_noise_input, 3, 1)
        layout.addWidget(self.l_noise_lbl2, 4, 0, 1, 2)
        layout.addWidget(self.update_btn, 5, 0, 1, 2, Qt.AlignHCenter)
        layout.addWidget(self.cancel_btn, 6, 0, 1, 2, Qt.AlignHCenter)

        # layout = QVBoxLayout()
        # layout.addWidget(self.method_lbl)
        # layout.addWidget(self.fir_rbtn)
        # layout.addWidget(self.iir_rbtn)
        # layout.addWidget(self.l_noise_lbl1)
        # layout.addWidget(self.l_noise_lbl2)
        # layout.addWidget(self.l_noise_input)
        # layout.addWidget(self.update_btn)
        # layout.addWidget(self.cancel_btn)
        self.setLayout(layout)

    def update(self):

        for r_btn in [self.fir_rbtn, self.iir_rbtn]:
            if r_btn.isChecked():
                method = r_btn.text().lower()

        freqs_lst = []  # empty list to hold all frequencies to be removed with notch filter

        try:
            if ", " in self.l_noise_input.text():
                split = self.l_noise_input.text().split(", ")
                for value in split:
                    freqs_lst.append(int(value))

            else:
                freqs_lst.append(int(self.l_noise_input.text()))

        except ValueError:
            QMessageBox.warning(self, "Invalid Input", "Enter one or more integer values eg 50, 100, 150")
            self.l_noise_input.clear()
            return

        freqs = tuple(freqs_lst)
        process_name = "line_noise"
        process_params = {"freqs": freqs, "method": method}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "Line noise removal settings updated")

        self.close()

class EpochOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(EpochOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        self.setWindowTitle("Epoch")

        self.fixed_length_rbtn = QRadioButton("Fixed length epochs")
        self.fixed_length_rbtn.clicked.connect(self.toggle_options)

        self.tstep_lbl = QLabel("Epoch interval")
        self.tstep_input = QDoubleSpinBox()
        self.tstep_input.setMinimum(0.1)
        self.tstep_input.setSingleStep(0.1)

        self.events_rbtn = QRadioButton("Epoch around events")
        self.events_rbtn.clicked.connect(self.toggle_options)

        self.tmin_lbl = QLabel("Tmin")
        self.tmin_input = QDoubleSpinBox()
        self.tmin_input.setMinimum(-10000)
        self.tmin_input.setSingleStep(0.1)

        self.tmax_lbl = QLabel("Tmax")
        self.tmax_input = QDoubleSpinBox()
        self.tmax_input.setMinimum(-10000)
        self.tmax_input.setSingleStep(0.1)

        self.autoreject_check = QCheckBox("Drop bad segments and epochs\nusing Autoreject thresholds")
        # if filter options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "epoch":

            if process_params["tmin"] != None: # ie if events data
                self.events_rbtn.setChecked(True)
                self.fixed_length_rbtn.setChecked(False)
                self.tmin_input.setValue(process_params["tmin"])
                self.tmax_input.setValue(process_params["tmax"])
                self.tstep_input.setValue(5.0) # default value

            elif process_params["tstep"] != None: # ie if resting state data
                self.events_rbtn.setChecked(False)
                self.fixed_length_rbtn.setChecked(True)
                self.tstep_input.setValue(process_params["tstep"])
                self.tmin_input.setValue(-0.2) # default value
                self.tmax_input.setValue(0.5) # default value

            self.toggle_options()

            self.autoreject_check.setChecked(process_params["autoreject"])

        else:# set default values
            self.tstep_input.setValue(5.0)
            self.tmin_input.setValue(-0.2)
            self.tmax_input.setValue(0.5)

            if self.main_window.resting_r_btn.isChecked():
                self.fixed_length_rbtn.setChecked(True)
                self.events_rbtn.setChecked(False)
            elif self.main_window.erp_r_btn.isChecked():
                self.fixed_length_rbtn.setChecked(False)
                self.events_rbtn.setChecked(True)
            self.toggle_options()

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        layout = QGridLayout()
        layout.addWidget(self.fixed_length_rbtn, 0, 0, 1, 2)
        layout.addWidget(self.tstep_lbl, 1, 0)
        layout.addWidget(self.tstep_input, 1, 1)

        layout.addWidget(self.events_rbtn, 2, 0, 1, 2)
        layout.addWidget(self.tmin_lbl, 3, 0)
        layout.addWidget(self.tmin_input, 3, 1)
        layout.addWidget(self.tmax_lbl, 4, 0)
        layout.addWidget(self.tmax_input, 4, 1)

        layout.addWidget(self.autoreject_check, 5, 0, 1, 2)
        layout.addWidget(self.update_btn, 6, 0, 1, 2, Qt.AlignHCenter)
        layout.addWidget(self.cancel_btn, 7, 0, 1, 2, Qt.AlignHCenter)
        self.setLayout(layout)

    def update(self):

        # error check for trying to epoch data with no events/stim channel around events
        if self.main_window.resting_r_btn.isChecked() & self.events_rbtn.isChecked():
            QMessageBox.warning(self, "No Stim Channel Selected", "Stim channel must be selected in order to epoch around events")
            return

        if self.events_rbtn.isChecked():
            tmin = self.tmin_input.value()
            tmax=self.tmax_input.value()
            tstep = None
        elif self.fixed_length_rbtn.isChecked():
            tmin = None
            tmax = None
            tstep = self.tstep_input.value()

        process_name = "epoch"
        process_params = {"tmin": tmin,
                          "tmax": tmax,
                          "tstep": tstep,
                          "autoreject": self.autoreject_check.isChecked()}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "Epoch settings updated")

        self.close()

    def toggle_options(self):
        if self.events_rbtn.isChecked():
            self.tmin_input.setDisabled(False)
            self.tmax_input.setDisabled(False)
            self.tstep_input.setDisabled(True)
        elif self.fixed_length_rbtn.isChecked():
            self.tmin_input.setDisabled(True)
            self.tmax_input.setDisabled(True)
            self.tstep_input.setDisabled(False)

class ChannelOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(ChannelOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        self.setWindowTitle("Bad Channels")

        info = main_window.data.info
        channel_types = [k.upper() for k in get_channel_type_constants().keys()]

        self.channels_view = QListWidget()
        self.channels_view.setSpacing(7)
        self.channels_view.setSelectionMode(QListWidget.ExtendedSelection)

        for index, ch in enumerate(info["chs"]):
            kind = channel_type(info, index).upper()
            name = ch["ch_name"]
            item = QListWidgetItem(name + "    " + kind)
            self.channels_view.addItem(item)
            if name in info["bads"]:
                item.setSelected(True)

        self.interp_check = QCheckBox("Interpolate bad channels")

        # if channel options have already been selected for this idx,
        # set interp channel checkbox to the value in the process_params dict
        # Like 'saving' the option you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "channels":
            if process_params["interpolate"] == True:
                self.interp_check.setChecked(True)
            elif process_params["interpolate"] == False:
                self.interp_check.setChecked(False)
        else:  # if setting options for the first time (default values)
            self.interp_check.setChecked(True)

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        layout = QVBoxLayout()
        layout.addWidget(self.channels_view)
        layout.addWidget(self.interp_check)
        layout.addWidget(self.update_btn)
        layout.addWidget(self.cancel_btn)
        layout.setAlignment(self.update_btn, Qt.AlignHCenter)
        layout.setAlignment(self.cancel_btn, Qt.AlignHCenter)
        self.setLayout(layout)

    def update(self):

        bads = []
        for item in self.channels_view.selectedItems():
            ch_name = item.text().split("   ")[0]
            bads.append(ch_name)

        process_name = "channels"
        process_params = {"bads": bads, "interpolate": self.interp_check.isChecked()}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "Bad channel settings updated")

        self.close()

class BaseLineOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(BaseLineOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        self.setWindowTitle("Baseline")

        self.baseline_lbl = QLabel("Interval over which to baseline:")
        self.a_lbl = QLabel("Start")
        self.a_input = QDoubleSpinBox()
        self.a_input.setMinimum(-10000)
        self.a_input.setSingleStep(0.1)
        self.b_lbl = QLabel("End")
        self.b_input = QDoubleSpinBox()
        self.b_input.setMinimum(-10000)
        self.b_input.setSingleStep(0.1)

        # if filter options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "baseline":
            interval = process_params["interval"] # eg (a, b)
            self.a_input.setValue(interval[0])
            self.b_input.setValue(interval[1])
        else:  # if setting filter options for the first time
            self.a_input.setValue(-0.2) # default value
            self.b_input.setValue(0)  # default value

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        layout = QGridLayout()
        layout.addWidget(self.baseline_lbl, 0, 0, 1, 2)
        layout.addWidget(self.a_lbl, 1, 0)
        layout.addWidget(self.a_input, 1, 1)
        layout.addWidget(self.b_lbl, 2, 0)
        layout.addWidget(self.b_input, 2, 1)
        layout.addWidget(self.update_btn, 3, 0, 1, 2, Qt.AlignHCenter)
        layout.addWidget(self.cancel_btn, 4, 0, 1, 2, Qt.AlignHCenter)
        self.setLayout(layout)

    def update(self):

        process_name = "baseline"
        process_params = {"interval": (self.a_input.value(), self.b_input.value())}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "Baseline settings updated")

        self.close()

class ICAOptionsDialog(QWidget):
    def __init__(self, main_window, idx):
        super(ICAOptionsDialog, self).__init__()

        self.main_window = main_window
        self.process_idx = idx

        self.setWindowTitle("ICA")

        ica_methods = ["FastICA", "Infomax", "Picard"]
        picks = mne.pick_types(self.main_window.data.info, meg=True, eeg=True, eog=False, ecg=False)
        # print(len(picks))
        num_eeg_and_meg_chs = len(picks)

        # --- the following code is taken from the mnelab sourcecode runicadialog.py---#

        self.method_select = QComboBox()
        self.method_select.addItems(ica_methods)
        self.method_select.currentIndexChanged.connect(self.toggle_method_options)

        self.extended_label = QLabel("Extended:")
        self.extended = QCheckBox()

        self.n_comps_lbl = QLabel("Set number of components by:")

        self.n_components_r_btn = QRadioButton("Number of components:")
        self.n_components_r_btn.clicked.connect(self.toggle_n_comp_options)

        self.n_components = QSpinBox()
        self.n_components.setRange(1, num_eeg_and_meg_chs)

        self.n_components_var_r_btn = QRadioButton("Cumulative explained variance:")
        self.n_components_var_r_btn.clicked.connect(self.toggle_n_comp_options)

        self.n_components_var = QDoubleSpinBox()
        self.n_components_var.setRange(0.01, 0.99)
        self.n_components_var.setSingleStep(0.01)

        self.n_comps_btns = QButtonGroup()
        self.n_comps_btns.addButton(self.n_components_r_btn)
        self.n_comps_btns.addButton(self.n_components_var_r_btn)

        self.exclude_bad_segments_lbl = QLabel("Exclude bad segments\n (determined by Autoreject threshold):")
        self.exclude_bad_segments = QCheckBox()

        self.rejection_method_lbl = QLabel("Method for rejecting ICA components:")
        self.manual_rbtn = QRadioButton("Manually")
        self.ecg_eog_rbtn = QRadioButton("Automatically using EOG and ECG")

        self.exclusion_methods_btns = QButtonGroup()
        self.exclusion_methods_btns.addButton(self.manual_rbtn)
        self.exclusion_methods_btns.addButton(self.ecg_eog_rbtn)

        self.update_btn = QPushButton('Update Settings', self)
        self.update_btn.setFixedWidth(150)
        self.update_btn.clicked.connect(self.update)
        self.cancel_btn = QPushButton('Cancel', self)
        self.cancel_btn.setFixedWidth(150)
        self.cancel_btn.clicked.connect(lambda: options_dialog_cancel(self, self.main_window))

        grid = QGridLayout()
        grid.addWidget(QLabel("Method:"), 0, 0)
        grid.addWidget(self.method_select, 0, 1)
        grid.addWidget(self.extended_label, 1, 0)
        grid.addWidget(self.extended, 1, 1)
        grid.addWidget(self.n_comps_lbl, 3, 0)
        grid.addWidget(self.n_components_r_btn, 4, 0)
        grid.addWidget(self.n_components, 4, 1)
        grid.addWidget(self.n_components_var_r_btn, 5, 0)
        grid.addWidget(self.n_components_var, 5, 1)
        grid.addWidget(self.exclude_bad_segments_lbl, 6, 0)
        grid.addWidget(self.exclude_bad_segments, 6, 1)
        grid.addWidget(self.rejection_method_lbl, 7, 0, 1, 2)
        grid.addWidget(self.manual_rbtn, 8, 0)
        grid.addWidget(self.ecg_eog_rbtn, 9, 0, 1, 2)
        grid.addWidget(self.update_btn, 10, 0, 1, 2, Qt.AlignHCenter)
        grid.addWidget(self.cancel_btn, 11, 0, 1, 2, Qt.AlignHCenter)
        self.setLayout(grid)

        self.toggle_method_options()

        # --- end of referenced code ---#

        # if filter options have already been selected for this idx,
        # set the widgets to the values in the process_params dict
        # Like 'saving' the options you selected

        process_name = main_window.pipeline[idx][0]
        process_params = main_window.pipeline[idx][1]

        if process_name == "ICA":
            # set method
            for i in range(len(ica_methods)):
                if process_params["ica_method"] == ica_methods[i].lower():
                    self.method_select.setCurrentIndex(i)
            # set "extended"
            self.extended.setChecked(process_params["extended"])
            # set n_components

            if process_params["n_components"] >= 1: # ie if num components
                self.n_components_r_btn.setChecked(True)
                self.n_components.setDisabled(False)
                self.n_components.setValue(process_params["n_components"])
                self.n_components_var_r_btn.setChecked(False)
                self.n_components_var.setValue(0.95)# default value
                self.n_components_var.setDisabled(True)

            elif process_params["n_components"] < 1:  # ie if variance explained
                self.n_components_var_r_btn.setChecked(True)
                self.n_components_var.setDisabled(False)
                self.n_components_var.setValue(process_params["n_components"])
                self.n_components.setDisabled(True)
                self.n_components.setValue(num_eeg_and_meg_chs) # default value
                self.n_components_r_btn.setChecked(False)

            self.toggle_n_comp_options()
            # set autoreject
            self.exclude_bad_segments.setChecked(process_params["autoreject"])
            # set rejection method
            if process_params["rejection_method"] == "manual":
                self.manual_rbtn.setChecked(True)
                self.ecg_eog_rbtn.setChecked(False)
            elif process_params["rejection_method"] == "ecg_eog":
                self.manual_rbtn.setChecked(False)
                self.ecg_eog_rbtn.setChecked(True)

        else: # set default values (if selecting options for the first time)
            self.method_select.setCurrentIndex(0)
            self.extended.setChecked(False)
            self.n_components_r_btn.setChecked(True)
            self.n_components.setValue(num_eeg_and_meg_chs)
            self.n_components_var.setDisabled(True)
            self.n_components_var.setValue(0.95)
            self.exclude_bad_segments.setChecked(True)
            self.manual_rbtn.setChecked(True)
            self.ecg_eog_rbtn.setChecked(False)

    def toggle_method_options(self):
        # --- this function is taken from the mnelab sourcecode runicadialog.py---#
        """Toggle extended options."""
        if self.method_select.currentText() == "Picard":  # enable extended and ortho
            self.extended_label.setEnabled(True)
            self.extended.setEnabled(True)

        elif self.method_select.currentText() == "Infomax":  # enable extended
            self.extended_label.setEnabled(True)
            self.extended.setEnabled(True)
        else:
            self.extended_label.setEnabled(False)
            self.extended.setChecked(False)
            self.extended.setEnabled(False)

    def toggle_n_comp_options(self):

        if self.n_components_r_btn.isChecked():
            self.n_components.setDisabled(False)
            self.n_components_var.setDisabled(True)

        elif self.n_components_var_r_btn.isChecked():
            self.n_components.setDisabled(True)
            self.n_components_var.setDisabled(False)

    def update(self):

        if self.manual_rbtn.isChecked():
            rej_method = "manual"
        elif self.ecg_eog_rbtn.isChecked():
            rej_method = "ecg_eog"

        if self.n_components_r_btn.isChecked():
            n_components = self.n_components.value()
        elif self.n_components_var_r_btn.isChecked():
            n_components = self.n_components_var.value()

        process_name = "ICA"
        process_params = {"n_components": n_components,
                          "ica_method": self.method_select.currentText().lower(),
                          "extended": self.extended.isChecked(),
                          "autoreject": self.exclude_bad_segments.isChecked(),
                          "rejection_method": rej_method}
        self.main_window.update_pipeline(self.process_idx, process_name, process_params)

        alert_msg(self, "Update Successful", "ICA settings updated")

        self.close()






import mne


def set_reference_func(data, process_params):

    data_ref_to_av = data.copy().set_eeg_reference(ref_channels=process_params["ref_channels"], projection=False)

    return data_ref_to_av

# ---- test -----------------------------------------------------
#
# import os
#
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# epochs = mne.Epochs(raw, events, verbose=False)
# evoked = epochs.average()
#
# process_params = {"ref_channels": "average"}
# from mne.io.pick import get_channel_type_constants
# channel_types = [k.upper() for k in get_channel_type_constants().keys()]
# if "EEG" in channel_types:
#     output = set_reference_func(evoked, process_params)
#     print(output.info)
#     print(type(output))
# else:
#     print("Error: No EEG channel data found")
#




import mne
from PyQt5.QtWidgets import *
from matplotlib import pyplot as plt
from mne.io.pick import get_channel_type_constants, channel_type

from plot_func_dialog_classes import *

class PlotWindow(QWidget):
    def __init__(self, main_window):
        super(PlotWindow, self).__init__()

        self.main_window = main_window

        # initialise lists to hold tuples of data and data names here, populate later in load_in_data function
        self.data_list = [(),()] # to hold data as read in
        self.picks_data_list = [(),()] # to hold data once channels have been selected
        #self.data1 = ""
        #self.data2 = ""

        self.setWindowTitle("Plot Options")
        self.title_lbl = QLabel("Select up to two FIF files to visualise.\n \n"
                                "Both files must contain the same channels and events,\n"
                                " e.g. the results of two different preprocessing pipelines \n"
                                "run on the same raw data. \n \n"
                                "If you only wish to visualise one file, please \n"
                                "select it as File 1, not File 2. \n")

        self.open1_btn = QPushButton("Load 1st Data File")
        self.open1_btn.clicked.connect(lambda: self.load_in_data(1))
        self.file1_lbl = QLabel("File 1: No file loaded yet ")

        self.open2_btn = QPushButton("Load 2nd Data File")
        self.open2_btn.clicked.connect(lambda: self.load_in_data(2))
        self.file2_lbl = QLabel("File 2: No file loaded yet")

        self.select_chs_lbl = QLabel("Select channels to plot:")
        self.select_chs_cbox = QComboBox()
        self.select_chs_cbox.activated[str].connect(self.update_channels)
        self.select_chs_cbox.setFixedWidth(150)

        self.time_series_tab = TimeSeriesTab(self)
        self.psd_tab = PSDTab(self)
        self.topo_epoch_evoked_tab = TopoEpochsEvokedTab(self)
        self.topo_epoch_evoked_tab.setDisabled(True)

        self.tabs = QTabWidget()
        self.tabs.addTab(self.time_series_tab, "Time Series")
        self.tabs.addTab(self.psd_tab, "PSD")
        self.tabs.addTab(self.topo_epoch_evoked_tab, "Epochs and Evoked Responses")

        layout = QGridLayout()
        layout.addWidget(self.title_lbl, 1, 0, 1, 3)

        layout.addWidget(self.open1_btn, 2, 0)
        layout.addWidget(self.file1_lbl, 3, 0, 1, 3)
        layout.addWidget(self.open2_btn, 4, 0)
        layout.addWidget(self.file2_lbl, 5, 0, 1, 3)
        layout.addWidget(self.select_chs_lbl, 6, 0)
        layout.addWidget(self.select_chs_cbox, 6, 1)
        layout.addWidget(self.tabs, 8, 0, 1, 3)
        self.setLayout(layout)

    def load_in_data(self, data_num):

        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        fileName, _ = QFileDialog.getOpenFileName(self, "Open...", "",
                                                  "(*.fif)", options=options)

        if fileName != "":
            try:
                fname = fileName.split("/")[-1]
                # load data
                if fileName.endswith("raw.fif"):
                    data = mne.io.read_raw_fif(fileName)
                    self.populate_data_and_widgets(data, fname, data_num)
                elif fileName.endswith("epo.fif"):
                    data = mne.read_epochs(fileName)
                    self.populate_data_and_widgets(data, fname, data_num)

            except ValueError:
                QMessageBox.warning(self, "Error", "Problem reading file in to MNE. Check file type is supported")
                return

    def populate_data_and_widgets(self, data, filename, data_num):
        if data_num == 1:
            self.data_list[0] = (data, filename)
            self.file1_lbl.setText("File 1: " + filename)

            self.select_chs_cbox.clear()
            self.select_chs_cbox.addItems(["All data channels", "Pick channels"])
            if "eeg" in data:
                self.select_chs_cbox.addItem("EEG only")
            if "meg" in data:
                self.select_chs_cbox.addItem("MEG only")

            if isinstance(data, mne.BaseEpochs):
                # populate topo_epoch_evoked_tab combobox with keys from event_id dict
                # both sets of data should have the same events if they are the results of different
                # pipelines run on the same data, so just take the event ids from the first set of data
                event_dict = data.event_id
                event_ids = [i for i in event_dict.keys()]
                self.topo_epoch_evoked_tab.populate_cbox(event_ids)

        elif data_num == 2:
            self.data_list[1] = (data, filename)
            self.file2_lbl.setText("File 2: " + filename)

        self.toggle_enabled_plots()

    def update_channels(self):

        if len(self.data_list[0]) == 0:  # if the first data file has not been loaded
            QMessageBox.warning(self, "File 1 Required", "Ensure a data file has been selected for File 1")
            return

        else:
            for i in range(len(self.data_list)):
                try:
                    if len(self.data_list[i]) != 0: # if data has been loaded
                        data = self.data_list[i][0]
                        if self.select_chs_cbox.currentText() == "All data channels":
                            self.picks_data_list[i] = data.load_data().copy().pick_types(meg=True, eeg=True, stim=False, eog=False)

                        elif self.select_chs_cbox.currentText() == "MEG only":
                            self.picks_data_list[i] = data.load_data().copy().pick_types(meg=True, eeg=False, stim=False, eog=False)

                        elif self.select_chs_cbox.currentText() == "EEG only":
                            self.picks_data_list[i] = data.load_data().copy().pick_types(meg=False, eeg=True, stim=False, eog=False)
                except ValueError: # if channel type present in file/data 1 but not file/data 2
                    QMessageBox.warning(self, "Channel Error", "Selected channels not present in both data files")
                    self.select_chs_cbox.setCurrentIndex(0) # reset to "all channels"
                    return

            if self.select_chs_cbox.currentText() == "Pick channels":
            # If picking specific channels to include.
            # Needs to be called outside the for loop so it is only called once, as the picked channels are applied to both data sets
            # Gets picks (using data set 1) in the following window. Once picks are selected,
            # the window calls update_channels_for_picks to update both data items in the picks_data_list.
                pick_channels_dialog = PickChannelsDialog(self)
                pick_channels_dialog.show()

    def update_channels_for_picks(self):

        for i in range(len(self.data_list)):
            if len(self.data_list[i]) != 0:

                try:
                    data = self.data_list[i][0]
                    self.picks_data_list[i] = data.load_data().copy().pick_types(meg=False, eeg=False, stim=False,
                                                                                 eog=False, include=self.picks)
                except ValueError:
                    QMessageBox.warning(self, "Channel Error", "Selected channels not present in both data files")
                    self.select_chs_cbox.setCurrentIndex(0)  # reset to "all channels"
                    return

    def toggle_enabled_plots(self):
        #enable the epoch_and_evoked tab/plot functions if:
        # 1) only file 1 is loaded and it is epoched data
        # 2) both files are loaded and both files are epoched data
        data_item1 = self.data_list[0]
        data_item2 = self.data_list[1]

        # if only data one is loaded and it is epoched data
        if len(data_item1) != 0 and isinstance(data_item1[0], mne.BaseEpochs) and len(data_item2) == 0:
            self.topo_epoch_evoked_tab.setDisabled(False)
        # if both data are epochs
        elif len(data_item1) != 0 and len(data_item2) != 0 and isinstance(data_item1[0], mne.BaseEpochs) and isinstance(data_item2[0], mne.BaseEpochs):
            self.topo_epoch_evoked_tab.setDisabled(False)
        else:
            self.topo_epoch_evoked_tab.setDisabled(True)

# ---- test ---------------------------------------------------------------
# import sys
# app = QApplication(sys.argv)
# main_window = QWidget()
# main_window.show()
# dialog = PlotWindow(main_window)
# dialog.show()
# sys.exit(app.exec_())




import mne
from PyQt5.QtWidgets import QMessageBox
from autoreject import get_rejection_threshold
from mne.preprocessing import ICA


# generation of ICA is the same process for raw and epoch
from check_channel_types import check_for_ch_type


def generate_ica(data, process_params, parent_window):

    print()
    print("------------------GENERATING ICA ---------------")
    print()

    # Set max_pca_components to the lesser value of its given argument or the rank
    # of the data, to avoid artifacts from incorrect data dimensionality
    n_components = process_params["n_components"]
    data_rank_dict = mne.compute_rank(data, info=data.info, verbose=False)
    data_rank = sum(data_rank_dict.values())
    max_pca_components = data_rank
    if n_components > max_pca_components:
        QMessageBox.warning(parent_window, "Updating Number of Components", "Number of components is greater than rank of data. Calculated rank (" + str(
            data_rank) + ") will be used as number of components instead")
        # print("Number of components is greater than rank of data. Calculated rank (" + str(
        #     data_rank) + ") will be used as max number of components")
        n_components = max_pca_components

    print()
    print("n components: ", n_components)
    ()

    # generate ICA
    method = process_params["ica_method"]
    fit_params = None
    if process_params["extended"]:
        fit_params = dict(extended=True)
    ica = mne.preprocessing.ICA(n_components=n_components, max_pca_components=max_pca_components, random_state=97,
                                method=method, fit_params=fit_params)
    return ica

def fit_ica_with_autoreject(ica, data):

    print()
    print("------------------FITTING ICA WITH AUTOREJECT---------------")
    print()

    if isinstance(data, mne.io.BaseRaw):
        # Autoreject only works on epoched data, so these steps are needed for raw data:
        # see autoreject docs for details: https://autoreject.github.io/faq.html
        tstep = 1.0
        events = mne.make_fixed_length_events(data, duration=tstep)
        epochs = mne.Epochs(data, events=events, tmin=0.0, tmax=tstep, baseline=(0, 0))
        reject = get_rejection_threshold(epochs)
        ica.fit(epochs, reject=reject, tstep=tstep)

    elif isinstance(data, mne.BaseEpochs):
        # drop bad channels to calculate rejection thresholds
        epochs_no_bads = data.copy().drop_channels(data.info["bads"])
        reject = get_rejection_threshold(epochs_no_bads)
        # fit ica with rejection thresholds to data with all channels
        ica.fit(data, reject=reject)

def fit_ica_no_autoreject(ica, data):

    print()
    print("------------------FITTING ICA ---------------")
    print()

    reject = None
    ica.fit(data, reject=reject)

def reject_eog_eeg_comps(ica, data):

    print()
    print("------------------ REJECTING ECG AND EOG COMPONENTS ---------------")
    print()

    # Select and remove components that match the EOG and ECG channels (ie blink and ECG artifact removal)
    ica.exclude = []
    eog_indices, eog_scores = ica.find_bads_eog(data)  # find which ICs match the EOG pattern
    ecg_indices, ecg_scores = ica.find_bads_ecg(data, method='correlation')

    ica.exclude += eog_indices
    ica.exclude += ecg_indices

    print("ica.exclude= ", ica.exclude)

def reject_comps_manual(ica, data):

    print()
    print("------------------ REJECTING COMPONENTS MANUALLY ---------------")
    print()

    print(ica.exclude)
    ica.plot_sources(data, block=True)  # plot time series of components - click on components to select them for exclusion at this point
    print(ica.exclude)

def apply_ica(ica, data):

    print()
    print("------------------ APPLYING ICA ---------------")
    print()

    reconst_data = data.copy()
    ica.apply(reconst_data)
    return reconst_data


# ---- test ------------------------------------------------------------------
#
# from PyQt5.QtWidgets import QMessageBox, QWidget, QApplication
# import sys
# app = QApplication(sys.argv)
# window = QWidget()
#
# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# # events = mne.find_events(raw, stim_channel='STI 014', verbose=False)
# # epochs = mne.Epochs(raw, events, verbose=False)
# data = raw
#
#
#
# process_params = {"n_components": 0.50,
#                   "ica_method": "fastica",
#                   "extended": False,
#                   #"ortho": self.ortho.isChecked(),
#                   "autoreject": False,
#                   #"rejection_method": "ecg_eog"}
#                   "rejection_method": "manual"}
#
#
# # only give the function for raw and epochs
# if isinstance(data, mne.io.BaseRaw) or isinstance(data, mne.BaseEpochs):
#
#     # 2) generate ICA
#     ica = generate_ica(data, process_params, window)
#
#     # 3) fit ICA (load data needed if epochs & dropping channels in autoreject)
#     #data.load_data()
#     if process_params["autoreject"]:
#         fit_ica_with_autoreject(ica, data)
#     else:
#         fit_ica_no_autoreject(ica, data)
#
#     # 4) Exclude components
#     if process_params["rejection_method"] == "manual":
#         reject_comps_manual(ica, data)
#
#     else: # if process_params["rejection_method"] == "ecg_eog"
#         # check that data actually has EOG and ECG/MEG channels
#         if check_for_ch_type(data, ["ECG", "EOG"]) \
#                 or check_for_ch_type(data, ["MAG", "EOG"]) \
#                 or check_for_ch_type(data, ["GRAD", "EOG"]):
#             reject_eog_eeg_comps(ica, data)
#         else:
#             print("Error: ", "EOG and/or ECG channels not found. Defaulting to manual rejection")
#             reject_comps_manual(ica, data)
#
#     # 5) Apply ICA
#     reconst_data = apply_ica(ica, data)
#
#     # 6) Plot data to check it
#
#     #raw - with and without ICA
#     ica.plot_sources(data)
#     ica.plot_overlay(data, exclude=ica.exclude, picks="eeg")
#     ica.plot_overlay(data, exclude=ica.exclude, picks="mag")

#     #epochs
#     # ica.plot_sources(data.average())
#     # ica.plot_overlay(data.average(), exclude=ica.exclude)
#
#
#
# #window.show()
# sys.exit(app.exec_())


import mne
from mne.io.pick import get_channel_type_constants, channel_type


def check_for_ch_type(data, required_chs):
    # takes an mne data object and a list of channel types eg ["MEG", "ECG"]
    # returns True if all specified channels are present in the data object
    # otherwise returns False

    all_channel_types = [k.upper() for k in get_channel_type_constants().keys()]

    # find all channel types that exist in data
    ch_types_in_data = []
    for index, ch in enumerate(data.info["chs"]):
        kind = channel_type(data.info, index).upper()
        if kind not in ch_types_in_data:
            ch_types_in_data.append(kind)

    # if all specified channel types are present in the list:
    if all(ch_type in all_channel_types for ch_type in required_chs):
        return True
    else:
        return False



import mne

# notch filter only works for raw data
# - error handling for this is implemented in the MainWindow so that return
#   can be used to leave the run_preprocessing function

def n_filter_func(data, process_params):
        data_notched = data.copy().notch_filter(freqs=process_params["freqs"])  # works with/without comma eg both freqs = (50) & (50,) are fine
        return data_notched


# ---- test -----------------------------------------------------

# import os
# sample_data_folder = mne.datasets.sample.data_path()
# sample_data_raw_file = os.path.join(sample_data_folder, "MEG", "sample", "sample_audvis_raw.fif")
#
# raw = mne.io.read_raw_fif(sample_data_raw_file, preload=True, verbose=False)
# process_params = {"freqs": (50, 100)}
# output = n_filter_func(raw, process_params)
# if output:
#     output.plot_psd()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mneUItest", # Replace with your own username
    version="0.0.15", # v0.0.14 wouldn't download
    author="Example Author",
    author_email="author@example.com",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    #package_dir={"": "src"},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "autoreject>=0.2",
        "matplotlib>=3.2",
        "mne>=0.20",
        "numpy>=1.19",
        "PyQt5>=5.10",
        "python-picard>=0.4",
        "QtPy>=1.9",
        "scikit-learn>=0.23",
        "scipy>=1.5"
    ],
    #py_modules=[
    #   "bad_chs",
    #    "baseline",
    #    "epochs",
    #    "filter",
    #    "ica",
    #    "load_data",
    #    "notch",
    #    "pipeline_dicts",
    #    "plot",
    #    "plot_func_dialog_classes",
    #    "prep_func_dialog_classes",
    #    "reference",
    #   "resample,"
    #    "save_data"
    #],
    python_requires='>=3.6',
)




import os
import sys
from PyQt5.QtGui import QIcon

from matplotlib import pyplot as plt

from plot import PlotWindow
from prep_func_dialog_classes import *
from save_data import get_valid_save_name, write_fif
from load_data import get_open_file_name, read_raw
from filter import filter_func
from resample import resample_func, resample_data_and_events_func
from notch import n_filter_func
from reference import set_reference_func
from baseline import baseline_func
from epochs import epoch_func, drop_bad_epochs_func
from bad_chs import mark_bad_chs, interp_bad_chs
from check_channel_types import check_for_ch_type
from ica import generate_ica, fit_ica_no_autoreject, fit_ica_with_autoreject, reject_eog_eeg_comps, apply_ica
from pipeline_dicts import get_preset_pipeline
from msgbxs import alert_msg

class MainWindow(QMainWindow):

    def __init__(self, app):
        super(MainWindow, self).__init__()

        self.app = app

        self.pipeline = [["", {}] for i in range(10)] # initialise

        self.options_windows = [QWidget for i in range(10)]

        process_names = ["None", "Filter", "Resample", "Remove Line Noise", "Set Reference (to Average)", "ICA",
                        "Mark Bad Channels", "Epoch", "Baseline Correct Epochs"] #, "Average Epochs"]

        self.btns = [QPushButton("Options", self) for i in range(10)]
        self.cBoxes = [QComboBox(self) for i in range(10)]

        for i in range(10):
            self.btns[i].setDisabled(True)
            self.cBoxes[i].setMaximumWidth(200)
            self.cBoxes[i].setDisabled(True)

        for i in range(10):
            self.cBoxes[i].addItems(process_names)
            #self.cBoxes[i].activated[str].connect(lambda: self.open_options(self.cBoxes[i].currentText(), i))

        self.cBoxes[0].activated[str].connect(lambda: self.open_options(self.cBoxes[0].currentText(), 0))
        self.cBoxes[1].activated[str].connect(lambda: self.open_options(self.cBoxes[1].currentText(), 1))
        self.cBoxes[2].activated[str].connect(lambda: self.open_options(self.cBoxes[2].currentText(), 2))
        self.cBoxes[3].activated[str].connect(lambda: self.open_options(self.cBoxes[3].currentText(), 3))
        self.cBoxes[4].activated[str].connect(lambda: self.open_options(self.cBoxes[4].currentText(), 4))
        self.cBoxes[5].activated[str].connect(lambda: self.open_options(self.cBoxes[5].currentText(), 5))
        self.cBoxes[6].activated[str].connect(lambda: self.open_options(self.cBoxes[6].currentText(), 6))
        self.cBoxes[7].activated[str].connect(lambda: self.open_options(self.cBoxes[7].currentText(), 7))
        self.cBoxes[8].activated[str].connect(lambda: self.open_options(self.cBoxes[8].currentText(), 8))
        self.cBoxes[9].activated[str].connect(lambda: self.open_options(self.cBoxes[9].currentText(), 9))

        self.btns[0].clicked.connect(lambda: self.open_options(self.cBoxes[0].currentText(), 0))
        self.btns[1].clicked.connect(lambda: self.open_options(self.cBoxes[1].currentText(), 1))
        self.btns[2].clicked.connect(lambda: self.open_options(self.cBoxes[2].currentText(), 2))
        self.btns[3].clicked.connect(lambda: self.open_options(self.cBoxes[3].currentText(), 3))
        self.btns[4].clicked.connect(lambda: self.open_options(self.cBoxes[4].currentText(), 4))
        self.btns[5].clicked.connect(lambda: self.open_options(self.cBoxes[5].currentText(), 5))
        self.btns[6].clicked.connect(lambda: self.open_options(self.cBoxes[6].currentText(), 6))
        self.btns[7].clicked.connect(lambda: self.open_options(self.cBoxes[7].currentText(), 7))
        self.btns[8].clicked.connect(lambda: self.open_options(self.cBoxes[8].currentText(), 8))
        self.btns[9].clicked.connect(lambda: self.open_options(self.cBoxes[9].currentText(), 9))

        self.fname_lbl = QLabel("Data file:")
        self.fname_lbl.setWordWrap(True)

        self.resting_r_btn = QRadioButton("No stimulus channel")
        self.resting_r_btn.clicked.connect(self.toggle_options)
        self.resting_r_btn.setDisabled(True)

        self.erp_r_btn = QRadioButton("Select stimulus channel")
        self.erp_r_btn.clicked.connect(self.toggle_options)
        self.erp_r_btn.setDisabled(True)

        self.stim_chan_select = QComboBox(self)
        self.stim_chan_select.activated[str].connect(lambda: self.get_stim_ch_and_events(self.data))
        self.stim_chan_select.setDisabled(True)
        self.stim_chan_select.setMinimumWidth(100)

        self.confirm_btn = QPushButton("Run pipeline and save results", self)
        self.confirm_btn.clicked.connect(self.run_preprocessing)
        self.confirm_btn.setDisabled(True)

        grid_layout = QGridLayout()

        grid_layout.addWidget(self.fname_lbl, 0, 0, 1, 2)
        grid_layout.addWidget(self.resting_r_btn, 1, 0)
        grid_layout.addWidget(self.erp_r_btn, 2, 0)
        grid_layout.addWidget(self.stim_chan_select, 2, 1)

        for i in range(10):
            grid_layout.addWidget(self.cBoxes[i], i+3, 0)
            grid_layout.addWidget(self.btns[i], i+3, 1)
        grid_layout.addWidget(self.confirm_btn,14, 0)

        #grid_layout.setColumnMinimumWidth(1, 100)

        #--------------------------

        openAct = QAction(QIcon('open.png'), '&Open...', self)
        openAct.triggered.connect(self.load_in_data)

        load_example_dataAct = QAction("&Download Example Data", self)
        load_example_dataAct.triggered.connect(self.load_example_data)

        plotAct = QAction('&Plot', self)
        plotAct.triggered.connect(self.plot_data)

        self.ex_pipelineAct = QAction('&Example ERP Pipeline', self)
        self.ex_pipelineAct.triggered.connect(lambda: self.load_preset_pipeline("example"))
        self.ex_pipelineAct.setDisabled(True)
        self.g_pipelineAct = QAction('&Gramfort ERP Pipeline', self)
        self.g_pipelineAct.triggered.connect(lambda: self.load_preset_pipeline("gramfort"))
        self.g_pipelineAct.setDisabled(True)
        self.m_pipelineAct = QAction('&Makoto ERP Pipeline', self)
        self.m_pipelineAct.triggered.connect(lambda: self.load_preset_pipeline("makoto"))
        self.m_pipelineAct.setDisabled(True)
        self.l_pipelineAct = QAction('&Luck ERP Pipeline', self)
        self.l_pipelineAct.triggered.connect(lambda: self.load_preset_pipeline("luck"))
        self.l_pipelineAct.setDisabled(True)
        self.r_pipelineAct = QAction('&Resting State Pipeline', self)
        self.r_pipelineAct.triggered.connect(lambda: self.load_preset_pipeline("resting"))
        self.r_pipelineAct.setDisabled(True)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(openAct)
        fileMenu.addAction(load_example_dataAct)
        plotMenu = menubar.addMenu("&Plot")
        plotMenu.addAction(plotAct)
        pipelineMenu = menubar.addMenu("&Load Pipeline")
        pipelineMenu.addAction(self.ex_pipelineAct)
        pipelineMenu.addAction(self.g_pipelineAct)
        pipelineMenu.addAction(self.m_pipelineAct)
        pipelineMenu.addAction(self.l_pipelineAct)
        pipelineMenu.addAction(self.r_pipelineAct)

        widget = QWidget()
        widget.setLayout(grid_layout)
        self.setCentralWidget(widget)
        self.setWindowTitle("Preprocessing Window")

    def load_in_data(self):
        try:
            fileName = get_open_file_name(self)
        except TypeError: # occurs if a file is double clicked rather than clicked on then "open" button clicked
            alert_msg(self, "Alert", "Please select file or directory then click open, rather than double clicking")
            return

        if fileName != "":
            self.fileName = fileName
            try:
                # load data
                self.data = read_raw(self.fileName)

                # set label to reflect opened file
                fname = fileName.split("/")[-1]
                self.fname_lbl.setText("Data File: " + fname)
                #resize the label widget dynamically to ensure it is large enough for the string
                new_size = self.fname_lbl.sizeHint()
                self.fname_lbl.setMinimumSize(new_size)

                # enable radio buttons for erp  vs resting data
                self.erp_r_btn.setDisabled(False)
                self.resting_r_btn.setDisabled(False)


            except ValueError:
                QMessageBox.warning(self, "Error", "Error reading file in to MNE. Check that file contains raw data "
                                                   "only and that file type is supported")
                return

    def load_example_data(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        dir_name = QFileDialog.getExistingDirectory(self, "Save example data in...", "", options=options)

        if dir_name != "":
            try:
                sample_data_folder = mne.datasets.sample.data_path(dir_name)
                sample_data_raw_file = os.path.join(sample_data_folder, 'MEG', 'sample',
                                                    'sample_audvis_raw.fif')
                fileName = sample_data_raw_file

                # load data
                self.data = read_raw(fileName)
                # set label to reflect opened file
                fname = fileName.split("/")[-1]
                self.fname_lbl.setText("Data File: " + fname)
                # resize the label widget dynamically to ensure it is large enough for the string
                new_size = self.fname_lbl.sizeHint()
                self.fname_lbl.setMinimumSize(new_size)

                # enable radio buttons for erp  vs resting data
                self.erp_r_btn.setDisabled(False)
                self.resting_r_btn.setDisabled(False)

            except ValueError:
                QMessageBox.warning(self, "Error", "Problem reading file in to MNE. Check file type is supported")
                return

    def get_stim_ch_and_events(self, data):

        data.stim_channel = self.stim_chan_select.currentText()
        print(data.stim_channel)
        data.events = mne.find_events(data, stim_channel=data.stim_channel)
        if len(data.events) == 0:  # channel chosen from stim channel does not contain events
            QMessageBox.warning(self, "Error", "Selected stim channel does not contain any events")
            return

        # once events have been successfully loaded, make rest of GUI live
        for i in range(10):
            self.btns[i].setDisabled(False)
            self.cBoxes[i].setDisabled(False)
        self.confirm_btn.setDisabled(False)
        self.ex_pipelineAct.setDisabled(False)
        self.g_pipelineAct.setDisabled(False)
        self.m_pipelineAct.setDisabled(False)
        self.l_pipelineAct.setDisabled(False)
        self.r_pipelineAct.setDisabled(False)

    def toggle_options(self):

        if self.erp_r_btn.isChecked():
            # enable the combobox to select stim channel and populate it with channel names
            # so that stim channel and events can be got
            self.stim_chan_select.setDisabled(False)
            self.stim_chan_select.clear()
            self.stim_chan_select.addItems(self.data.info["ch_names"])

            # disable the rest of the GUI until events are successfully loaded
            for i in range(10):
                self.btns[i].setDisabled(True)
                self.cBoxes[i].setDisabled(True)
            self.confirm_btn.setDisabled(True)
            self.r_pipelineAct.setDisabled(True)

        if self.resting_r_btn.isChecked():
            # resting state data, therefore set stim_channel and events to empty list
            self.stim_chan_select.setDisabled(True)
            self.data.stim_channel = None
            self.data.events = []

            # disable ERP preset pipelines
            self.ex_pipelineAct.setDisabled(True)
            self.g_pipelineAct.setDisabled(True)
            self.m_pipelineAct.setDisabled(True)
            self.l_pipelineAct.setDisabled(True)

            # make rest of GUI live
            for i in range(10):
                self.btns[i].setDisabled(False)
                self.cBoxes[i].setDisabled(False)
            self.confirm_btn.setDisabled(False)
            self.r_pipelineAct.setDisabled(False) # enable resting state preset pipeline

    def open_options(self, str, idx):
        if str == "None":
            self.pipeline[idx] = ["", {}] # reset in case a process was chosen then reset to None
            alert_msg(self, "Options", "No options")
            #print(self.pipeline)
            return
        elif str == "Filter":
            self.options_windows[idx] = FilterOptionsDialog(self, idx)
        elif str == "Resample":
            self.options_windows[idx] = ResampleOptionsDialog(self, idx)
        elif str == "Mark Bad Channels": # remove/interpolate
            self.options_windows[idx] = ChannelOptionsDialog(self, idx)
        elif str == "Set Reference (to Average)":
            # No options therefore no options window needed
            self.pipeline[idx] = ["set_reference", {"ref_channels": "average"}]
            alert_msg(self, "Set Reference Options", "No options required for this process")
            return
        elif str == "Remove Line Noise":
            self.options_windows[idx] = LineNoiseOptionsDialog(self, idx)
        elif str == "Epoch":
            self.options_windows[idx] = EpochOptionsDialog(self, idx)
        elif str == "ICA":
            self.options_windows[idx] = ICAOptionsDialog(self, idx)
        elif str == "Baseline Correct Epochs":
            self.options_windows[idx] = BaseLineOptionsDialog(self, idx)
        self.options_windows[idx].show()

    def load_preset_pipeline(self, pipeline_name):

        self.pipeline = get_preset_pipeline(pipeline_name, self.data)

        process_list_idx_lookup = {
            "": 0, #
            "filter": 1, #
            "resample": 2, #
            "line_noise": 3, #
            "set_reference": 4, #
            "ICA": 5,
            "channels": 6, #
            "epoch": 7, # but still needs testing, doesn't always work (memory))
            "baseline": 8 #
        }
        for i in range(len(self.pipeline)):
            process_idx = process_list_idx_lookup[self.pipeline[i][0]] # look up process name
            self.cBoxes[i].setCurrentIndex(process_idx)
            #
            if self.pipeline[i][0] == "channels":
                # to get bad chs before running the pipeline
                # all other params are supplied by the params dict or by default values
                self.open_options(self.cBoxes[i].currentText(), i)

    def update_pipeline(self, idx, process, process_params):
        self.pipeline[idx][0] = process
        self.pipeline[idx][1] = process_params
        print(self.pipeline)

    def get_pipeline_process_name(self, idx):
        return self.pipeline[idx][0]

    def get_pipeline_process_params(self, idx):
        return self.pipeline[idx][1]

    def run_preprocessing(self):

        print(self.pipeline)

        # # get filename to save file once processing is done
        save_name = get_valid_save_name(self)
        if save_name == -1: # if no valid save name
            return

        # if valid save name, do preprocessing

        self.statusBar().showMessage('Running pipeline...')
        self.app.processEvents() # use this to update the GUI while the function is running. Aware not a good solution and
        # best practice would likely be to run the function in a worker thread, but implementing as a quick fix to get the
        # functionality out for testing

        # at this point, if the data is ERP data the events have been saved as raw.events
        # if the data does not have events/a stim channel, raw.events will be an empty list

        #processes_list = [raw] # hold the resulting data after each process/step, so that each step can access the most recently processed data
        data = self.data.copy() # start with direct copy, overwrite it at each step (preserves more memory than keeping all th individual steps in a list)
        for i in range(len(self.pipeline)):

            if self.pipeline[i] != ["", {}]:
                #data = processes_list[-1]]
                if isinstance(data, mne.io.BaseRaw) or isinstance(data, mne.BaseEpochs):
                    data.load_data() # error if trying to load evoked data(?)
                process_name = self.get_pipeline_process_name(i)
                process_params = self.get_pipeline_process_params(i)
                if process_name == "filter":
                    self.statusBar().showMessage('Applying filter...')
                    self.app.processEvents()

                    #data_filtered = data.copy().filter(l_freq = process_params["l_freq"], h_freq = process_params["h_freq"], method=process_params["method"])
                    data_filtered = filter_func(data, process_params)
                    data = data_filtered # overwrite
                    del data_filtered # delete to free up RAM
                    #processes_list.append(data_filtered)
                    print("filtering finished")
                    print()
                elif process_name == "resample":
                    self.statusBar().showMessage('Resampling...')
                    self.app.processEvents()

                    #data_resampled = data.copy().resample(sfreq=process_params["sfreq"])
                    if len(data.events) == 0: # no events/stim channel
                        data_resampled = resample_func(data, process_params)
                    else: # events need to be resampled as well as data
                        data_resampled, events_resampled = resample_data_and_events_func(data, process_params)
                        data_resampled.events = events_resampled # overwrite existing events
                    #processes_list.append(data_resampled)
                    data = data_resampled # overwrite
                    del data_resampled # delete to free up RAM
                    print("resampling finished")
                    print()
                elif process_name == "line_noise":
                    self.statusBar().showMessage('Removing line noise...')
                    self.app.processEvents()

                    if isinstance(data, mne.io.BaseRaw):
                    #data_notched = data.copy().notch_filter(freqs=process_params["freqs"]) # works with/without comma eg both freqs = (50) & (50,) are fine
                        data_notched = n_filter_func(data, process_params)
                        #processes_list.append(data_notched)
                        data = data_notched  # overwrite
                        del data_notched  # delete to free up RAM
                        print("notch filter finished")
                        print()
                    else:
                        QMessageBox.warning(self, "Error", "Notch filter can only be applied to raw data")
                        self.statusBar().showMessage("")
                        return
                elif process_name == "set_reference":
                    self.statusBar().showMessage('Setting EEG reference...')
                    self.app.processEvents()

                    # check that data actually has EEG channels
                    if check_for_ch_type(data, ["EEG"]):
                        #data_ref_to_av = data.copy().set_eeg_reference(ref_channels=process_params["ref_channels"], projection=False)
                        data_ref_to_av = set_reference_func(data, process_params)
                        #processes_list.append(data_ref_to_av)
                        data = data_ref_to_av  # overwrite
                        del data_ref_to_av  # delete to free up RAM
                        print("set reference finished")
                        print()

                    else:
                        QMessageBox.warning(self, "Error", "No EEG channels found")
                        self.statusBar().showMessage("")
                        return
                elif process_name == "channels":
                    self.statusBar().showMessage('Marking bad channels...')
                    self.app.processEvents()
                    data_bad_chs = mark_bad_chs(data, process_params)
                    if process_params["interpolate"]:
                        self.statusBar().showMessage('Interpolating bad channels...')
                        self.app.processEvents()
                        #data.plot(block=True) # plots to check that the channels (in raw data) have been interpolated
                        data_bads_interp = interp_bad_chs(data_bad_chs)
                        #data_bads_interp.plot(block=True) # plots to check that the channels have been interpolated
                        print("bad channels interpolation finished. interpolated:  ", data_bads_interp.info["bads"])
                        # processes_list.append(data_bads_interp)
                        data = data_bads_interp  # overwrite
                        del data_bads_interp  # delete to free up RAM
                    else:
                        #processes_list.append(data_bad_chs)
                        print("bad channels finished: ", data_bad_chs.info["bads"])
                        data = data_bad_chs  # overwrite
                    del data_bad_chs  # delete to free up RAM
                elif process_name == "epoch":

                    if isinstance(data, mne.io.BaseRaw):
                        #try:
                        if process_params["autoreject"]:
                            self.statusBar().showMessage('Epoching (and dropping bad epochs)...')
                            self.app.processEvents()
                            epochs_drop_bad = drop_bad_epochs_func(data, process_params)
                            # processes_list.append(epochs_drop_bad)
                            data = epochs_drop_bad  # overwrite
                            del epochs_drop_bad  # delete to free up RAM
                        else: # if no autoreject
                            self.statusBar().showMessage('Epoching...')
                            self.app.processEvents()
                            epochs = epoch_func(data, process_params)
                            #processes_list.append(epochs)
                            data = epochs  # overwrite
                            del epochs  # delete to free up RAM
                        print("epoching finished")
                        print()
                    else:
                        QMessageBox.warning(self, "Error", "Epoching can only be applied to raw data")
                        self.statusBar().showMessage("")
                        return
                elif process_name == "baseline":
                    self.statusBar().showMessage('Applying baseline...')
                    self.app.processEvents()
                    try:
                        # epochs_baselined = data.copy().apply_baseline(process_params["interval"])
                        epochs_baselined = baseline_func(data, process_params)
                    except ValueError as VE: # if baseline interval is outside of epoch interval limits
                        QMessageBox.warning(self, "Error", str(VE))
                        self.statusBar().showMessage("")
                        return
                    except AttributeError as AE: # if data other than epoch or evoked
                        QMessageBox.warning(self, "Error", "Baseline can only be applied to epoched or evoked data")
                        self.statusBar().showMessage("")
                        return
                    #processes_list.append(epochs_baselined)
                    data = epochs_baselined  # overwrite
                    del epochs_baselined  # delete to free up RAM
                    print("baselining finished")
                    print()
                elif process_name == "ICA":
                    self.statusBar().showMessage('Performing ICA (may take a while)...')
                    self.app.processEvents()
                    if isinstance(data, mne.io.BaseRaw) or isinstance(data, mne.BaseEpochs):

                        # 1) generate ICA
                        self.statusBar().showMessage('Generating ICA (may take a while)...')
                        self.app.processEvents()
                        self.ica = generate_ica(data, process_params, self)

                        # 2) fit ICA
                        self.statusBar().showMessage('Fitting ICA (may take a while)...')
                        self.app.processEvents()
                        if process_params["autoreject"]:
                            fit_ica_with_autoreject(self.ica, data)
                        else:
                            fit_ica_no_autoreject(self.ica, data)

                        # 3) exclude components
                        self.statusBar().showMessage('Excluding components (may take a while)...')
                        self.app.processEvents()
                        if process_params["rejection_method"] == "ecg_eog":
                            # check that data actually has EOG and ECG/MEG channels
                            if check_for_ch_type(data, ["ECG", "EOG"]) \
                                    or check_for_ch_type(data, ["MAG", "EOG"]) \
                                    or check_for_ch_type(data, ["GRAD", "EOG"]):
                                reject_eog_eeg_comps(self.ica, data)
                            else:
                                QMessageBox.warning(self, "Automated Rejection Error",
                                                    "EOG and/or ECG channels not found. Defaulting to manual rejection")
                        else: # if manual exclusion
                            # cannot use reject_comps_manual function here because
                            # the plot does not show when called in another module
                            print()
                            print("------------------ REJECTING COMPONENTS MANUALLY ---------------")
                            print()
                            print(self.ica.exclude)

                            fig = self.ica.plot_sources(data)  # plot time series of components
                            fig.subplots_adjust(top=0.9)
                            plt.suptitle("1) Select components to be removed by clicking on their time series. \n"
                                         "Clicking on the name of each component will show its topographic plot. \n"
                                         "2) To continue, press enter or close the plot", fontsize=12)

                            # keep the figure open for selection of components to remove,
                            # until the plot is exited or enter is pressed

                            self.ica.finished = False

                            def press(event):
                                print('press', event.key)
                                sys.stdout.flush()
                                if event.key == 'enter':
                                    self.ica.finished = True

                            fig.canvas.mpl_connect('key_press_event', press)

                            while self.ica.finished != True & plt.fignum_exists(fig.number):
                                plt.pause(0.1)
                            plt.close()

                            # the following works to have the plot pause until it is exited, but not if enter is pressed
                            # (Keeping the press enter to continue as it is more intuitive for users)
                            # while plt.fignum_exists(fig.number):
                            #   plt.pause(0.1)

                            print(self.ica.exclude)

                            #
                            #self.ica.plot_overlay(data, exclude=self.ica.exclude, picks="eeg")
                            #self.ica.plot_overlay(data, exclude=self.ica.exclude, picks="mag")

                        # 4) apply ICA
                        self.statusBar().showMessage('Applying ICA...')
                        self.app.processEvents()
                        reconst_data = apply_ica(self.ica, data)


                        # 5) save results
                        #processes_list.append(reconst_data)
                        data = reconst_data  # overwrite
                        del reconst_data  # delete to free up RAM
                        print("ICA finished")
                        print()

                    else:
                        QMessageBox.warning(self, "Error", "ICA is currently only available for raw or epoched data types.")
                        self.statusBar().showMessage("")
                        return

        self.statusBar().showMessage('Saving...')
        print("Saving...")
        #saved = write_fif(processes_list[-1], save_name)
        saved = write_fif(data, save_name)
        if saved:
            alert_msg(self, "Pipeline Successful", "Pipeline run and data saved successfully")
        else:
            alert_msg(self, "Save Unsuccessful", "Problem saving data")

        self.statusBar().clearMessage() # clear the status bar now that pipeline has finished

    def plot_data(self):

        plot_dialog = PlotWindow(self)
        plot_dialog.show()

if __name__ == '__main__':
    app = QApplication(sys.argv)
    main_window = MainWindow()
    main_window.show()
    sys.exit(app.exec_())




import os
from pathlib import Path

import mne
from PyQt5.QtWidgets import *

# subclass of QFileDialog to allow opening of either a file or a directory
# Code for this class taken from StackOverflow, Luke's answer (05/07/11)
# https://stackoverflow.com/questions/6484793/multiple-files-and-folder-selection-in-a-qfiledialog?noredirect=1&lq=1
# Accessed 27/07/20

class OpenFileOrDirDialog(QFileDialog):
    def __init__(self,*args):
        QFileDialog.__init__(self,*args)
        self.setOptions(QFileDialog.DontUseNativeDialog)
        self.setFileMode(QFileDialog.ExistingFile) # use QFileDialog.ExstingFiles if wanting to open multiple files/dirs
        buttons = self.findChildren(QPushButton)
        for button in buttons:
            if 'open' in str(button.text()).lower():
                self.openFileButton=button
        self.openFileButton.clicked.disconnect()
        self.openFileButton.clicked.connect(self.openClicked)
        self.tree= self.findChild(QTreeView)

    def openClicked(self):

        inds = self.tree.selectionModel().selectedIndexes()
        files = []
        for i in inds:
            if i.column()==0:
                files.append(os.path.join(str(self.directory().absolutePath()),str(i.data())))
        self.selectedFiles = files
        self.hide()


    def filesSelected(self):
        return self.selectedFiles[0]


def get_open_file_name(main_window):
    open_file_dialog = OpenFileOrDirDialog(main_window, "Open...")
    open_file_dialog.exec_()
    fpath = open_file_dialog.filesSelected()
    return fpath

# Following read_raw function and supported dict taken/modified from mnelab code io readers.py
# Authors: Clemens Brunner <clemens.brunner@gmail.com>
# License: BSD (3-clause)

# supported read file formats
supported = {".edf": mne.io.read_raw_edf,
             ".bdf": mne.io.read_raw_bdf,
             ".gdf": mne.io.read_raw_gdf,
             ".vhdr": mne.io.read_raw_brainvision,
             ".fif": mne.io.read_raw_fif,
             ".fif.gz": mne.io.read_raw_fif,
             ".set": mne.io.read_raw_eeglab,
             ".cnt": mne.io.read_raw_cnt,
             ".mff": mne.io.read_raw_egi,
             ".nxe": mne.io.read_raw_eximia,
             ".hdr": mne.io.read_raw_nirx,
             ".ds": mne.io.read_raw_ctf} # added as most CUBRIC users use .ds format


def read_raw(fname, *args, **kwargs):
    """Read raw file.

    Parameters
    ----------
    fname : str
        File name to load.

    Returns
    -------
    raw : mne.io.Raw
        Raw object.

    Notes
    -----
    This function supports reading different file formats. It uses the supported
    dict to dispatch the appropriate read function for a supported file type.
    """
    ext = "".join(Path(fname).suffixes)
    if ext in supported:
        return supported[ext](fname, *args, **kwargs)




# ---- test -------------------------------------------------------------

#import sys
# app = QApplication(sys.argv)
# main_window = QWidget()
# main_window.show()
#
# fileName = get_open_file_name(main_window)
# if fileName != "":
#     main_window.fileName = fileName
#     main_window.data = read_raw(main_window.fileName)
#
# sys.exit(app.exec_())